package tools.descartes.prisma.agent.wildfly.callstack.impl;

import java.util.Collection;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicLong;

import tools.descartes.prisma.agent.wildfly.callstack.ComponentRecord;
import tools.descartes.prisma.agent.wildfly.callstack.InterfaceRecord;
import tools.descartes.prisma.agent.wildfly.callstack.InvocationGroupRecord;
import tools.descartes.prisma.agent.wildfly.callstack.OperationRecord;
import tools.descartes.prisma.agent.wildfly.callstack.TraceVisitor;
import tools.descartes.prisma.agent.wildfly.service.ModelExtractionService;

public class DefaultInvocationGroupRecord extends InvocationGroupRecord {
	private final AtomicLong invocationCount = new AtomicLong();
	private final AtomicLong executionTime = new AtomicLong();
	private final AtomicLong cpuTime = new AtomicLong();
	private final AtomicLong beforeCpuTime = new AtomicLong();
	private final ConcurrentHashMap<String, ComponentRecord> requiredComponents = new ConcurrentHashMap<String, ComponentRecord>();
	
	public DefaultInvocationGroupRecord(OperationRecord operation, String name) {
		super(operation, name);
	}
	
	public long getInvocationCount() {
		return invocationCount.get();
	}
	
	public long getExecutionTime() {
		return executionTime.get();
	}
	
	public String getAdditionalInformation() {
		return String.format("%d, %d, %d", executionTime.get() - getExecutionTimeOfCalls(), cpuTime.get(), beforeCpuTime.get());
	}
	
	public void addInvocation(long timeInNanos, long cpuTimeInNanos, long beforeCpuTimeInNanos) {
		invocationCount.incrementAndGet();
		executionTime.addAndGet(timeInNanos);
		cpuTime.addAndGet(cpuTimeInNanos);
		beforeCpuTime.addAndGet(beforeCpuTimeInNanos);
	}
	
	private long getExecutionTimeOfCalls() {
		long total = 0;
		for (ComponentRecord reqCrec : requiredComponents.values()) {
			for (InterfaceRecord reqIrec : reqCrec.getInterfaces()) {
				for (OperationRecord reqOrec : reqIrec.getOperations()) {
					double sum = 0.0;
					long totalWheight = 0;
					for (InvocationGroupRecord reqGrec : reqOrec.getInvocationGroups()) {
						if (reqGrec != null) {
							long count =  reqGrec.getInvocationCount();
							sum += count * reqGrec.getExecutionTime();
							totalWheight += count;
						}
					}
					if (totalWheight > 0) {
						total += (sum / totalWheight);
					}
					
				}
			}
		}
		return total;
	}
	
	public ComponentRecord requireComponent(String component) {
		ComponentRecord record = requiredComponents.get(component);
		if (record == null) {
			record = getTraceFactory().createComponentRecord(component, this);
			if (requiredComponents.putIfAbsent(component, record) != null) {
				// a different thread has changed it in the mean time
				record = requiredComponents.get(component);
			} else {
				ModelExtractionService.INSTANCE.newComponentRecord(record);
			}
		}
		return record;
	}
	
	public Collection<ComponentRecord> getRequiredComponents() {
		return requiredComponents.values();
	}
	
	public void accept(TraceVisitor visitor, InvocationGroupRecord source) {
//		System.out.println("Visit invocation group " + getName());
		if (source != null) {
			visitor.visitRequiredInvocationGroup(source, this);
		} else {
			visitor.visitInvocationGroup(this);
		}

		for (ComponentRecord reqCrec : requiredComponents.values()) {
			reqCrec.accept(visitor, this);
		}
	}

	@Override
	public boolean isFiltered() {
		return false;
	}

}
