package tools.descartes.prisma.agent.wildfly.interceptors;

import java.lang.management.ManagementFactory;
import java.lang.management.ThreadMXBean;
import java.util.Deque;

import javax.jms.Destination;
import javax.jms.Message;
import javax.jms.Queue;
import javax.jms.Topic;

import org.jboss.as.ejb3.component.EJBComponent;
import org.jboss.as.ejb3.component.interceptors.AbstractEJBInterceptor;
import org.jboss.as.ejb3.component.messagedriven.MessageDrivenComponent;
import org.jboss.invocation.InterceptorContext;
import org.jboss.logging.Logger;

import tools.descartes.prisma.agent.wildfly.callstack.ComponentRecord;
import tools.descartes.prisma.agent.wildfly.callstack.InterfaceRecord;
import tools.descartes.prisma.agent.wildfly.callstack.InvocationGroupRecord;
import tools.descartes.prisma.agent.wildfly.callstack.OperationRecord;
import tools.descartes.prisma.agent.wildfly.callstack.ThreadCallStack;
import tools.descartes.prisma.agent.wildfly.service.MonitoringService;

public class EJBIncomingCallInterceptor extends AbstractEJBInterceptor {
	
	private static final Logger log = Logger.getLogger(EJBIncomingCallInterceptor.class);
	
	private final MonitoringService monitoringService;
	
	private final ThreadMXBean threadBean;
		
	public EJBIncomingCallInterceptor(MonitoringService monitoringService) {
		this.monitoringService = monitoringService;
		this.threadBean = ManagementFactory.getThreadMXBean();
		if (this.threadBean.isCurrentThreadCpuTimeSupported()) {
			this.threadBean.setThreadCpuTimeEnabled(true);
		}
	}

	@Override
	public Object processInvocation(InterceptorContext ctx) throws Exception {		
		Deque<InvocationGroupRecord> threadCallStack = ThreadCallStack.get();
		EJBComponent ejbComponent = getComponent(ctx, EJBComponent.class);
		String component = ejbComponent.getApplicationName() + "/" + ejbComponent.getModuleName() + "/" + ejbComponent.getComponentName();
		String componentClass = ejbComponent.getComponentClass().getName();	
		String interf = ctx.getMethod().getDeclaringClass().getName();
		String operation = ctx.getMethod().getName();
		InvocationGroupRecord caller = threadCallStack.peek();
		ComponentRecord crec = monitoringService.getTraceRoot().requireComponent(component);
		if (crec.getComponentClass() == null) {
			crec.setComponentClass(componentClass);
		}
		
		if (ejbComponent instanceof MessageDrivenComponent
				&& operation.equals("onMessage")) {
			interf = "javax.jms.MessageListener";
			Message msg = (Message)ctx.getParameters()[0];
			Destination dest = msg.getJMSDestination();
			if (dest != null) {
				String destName = null;
				String destClass = null;
				if (dest instanceof Queue) {
					destName = ((Queue)dest).getQueueName();
					destClass = "javax.jms.Queue";
				} else if (dest instanceof Topic) {
					destName = ((Topic)dest).getTopicName();
					destClass = "javax.jms.Topic";
				}
				ComponentRecord destCrec = monitoringService.getTraceRoot().requireComponent(destName);
				if (destCrec.getComponentClass() == null) {
					destCrec.setComponentClass(destClass);
				}
				InvocationGroupRecord destGrec = destCrec.requireInterface("javax.jms.MessageProducer").requireOperation("send").requireInvocationGroup(0);
				InvocationGroupRecord destCalledGrec = getInvocationGroup(destGrec.requireComponent(component).requireInterface(interf).requireOperation(operation), ctx);
				destCalledGrec.addInvocation(0, 0, 0);
			}
		}
		
		InterfaceRecord	irec = crec.requireInterface(interf);
		OperationRecord orec = irec.requireOperation(operation);
		InvocationGroupRecord grec = getInvocationGroup(orec, ctx);
		threadCallStack.push(grec);
		
//		long startCpuTime = threadBean.getCurrentThreadCpuTime();
		long startTime = System.nanoTime();
		try {
			return ctx.proceed();
		} finally {
			long endTime = System.nanoTime();
//			long endCpuTime = threadBean.getCurrentThreadCpuTime();
			
			long executionTime = (endTime - startTime);
//			long cpuTime = (endCpuTime - startCpuTime);
//			long lastCpuTime = monitoringService.getLastCpuTime(Thread.currentThread().getId());
//			long beforeCpuTime = (lastCpuTime > 0) ? startCpuTime - lastCpuTime : 0;				
			if (caller != null) {
				// this is a local call in the same thread
				// also add a outgoing call to the callee component
				if (log.isTraceEnabled()) {
					log.info("Outgoing invocation (" + caller.getOperation().getInterface().getInterfaceName() + ", " + caller.getOperation().getOperationName() + ") -> (" + interf + ", " + operation + ", " + executionTime + "ms)");
				}
				ComponentRecord callerCrec = caller.requireComponent(component);
				callerCrec.setComponentClass(componentClass);
				OperationRecord callerOrec = callerCrec.requireInterface(interf).requireOperation(operation);
				InvocationGroupRecord callerGrec = getInvocationGroup(callerOrec, ctx);
				callerGrec.addInvocation(executionTime, 0, 0);
			}
			if (log.isTraceEnabled()) {
				log.info("Incoming invocation (" + component + ", " + operation + ", " + executionTime + "ms)");
			}
				
//			monitoringService.setLastCpuTime(Thread.currentThread().getId(), endCpuTime);
			grec.addInvocation(executionTime, 0, 0);
			threadCallStack.pop();
		}
	}
	
	private InvocationGroupRecord getInvocationGroup(OperationRecord orec, InterceptorContext ctx) {
		int[] names = orec.getObservedParameterIndeces();
		InvocationGroupRecord grec = null;
		if (names.length > 0) {
			int idx = orec.characterizeValue(0, ctx.getParameters()[names[0]]);
			grec = orec.requireInvocationGroup(idx);
		} else {
			grec = orec.requireInvocationGroup(0);
		}
		return grec;
	}
	
}
