package tools.descartes.prisma.core.adapter;

import static org.fest.assertions.api.Assertions.assertThat;
import static tools.descartes.prisma.core.util.DMLHelper.componentInternalBehavior;
import static tools.descartes.prisma.core.util.DMLHelper.internalAction;
import static tools.descartes.prisma.core.util.DMLHelper.resourceDemand;

import java.util.Collections;

import org.junit.Before;
import org.junit.Test;

import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.AbstractAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.BranchAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ComponentInternalBehavior;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.InternalAction;
import tools.descartes.prisma.core.ModelRepository;
import tools.descartes.prisma.core.TestModelRepository;
import tools.descartes.prisma.core.Transaction;
import tools.descartes.prisma.core.merge.ComponentInternalBehaviorMerger;

public class ComponentInternalBehaviorMergerTest {

	private ModelRepository repository;
	private Transaction transaction;

	@Before
	public void setUp() {
		repository = new TestModelRepository();
		transaction = repository.createTransaction();
	}

	@Test
	public void testMergeBothEmpty() {
		ComponentInternalBehavior baseBehavior = componentInternalBehavior();
		ComponentInternalBehavior otherBehavior = componentInternalBehavior();

		ComponentInternalBehaviorMerger merger = new ComponentInternalBehaviorMerger(repository);
		merger.mergeTree(transaction, baseBehavior, otherBehavior, Collections.emptySet());
		assertThat(baseBehavior.getActions()).isEmpty();
	}

	@Test
	public void testMergeOtherEmpty() {
		/*
		 * @formatter:off
		 */
		ComponentInternalBehavior baseBehavior = componentInternalBehavior(
			internalAction(resourceDemand("Test"))
			);
		ComponentInternalBehavior otherBehavior = componentInternalBehavior();
		/*
		 * @formatter:on
		 */
		ComponentInternalBehaviorMerger merger = new ComponentInternalBehaviorMerger(repository);
		merger.mergeTree(transaction, baseBehavior, otherBehavior, Collections.emptySet());
		assertThat(baseBehavior.getActions()).hasSize(1);
		assertThat(baseBehavior.getActions().get(0)).isInstanceOf(BranchAction.class);
	}

	@Test
	public void testMergeBaseEmpty() {
		/*
		 * @formatter:off
		 */
		ComponentInternalBehavior baseBehavior = componentInternalBehavior();
		ComponentInternalBehavior otherBehavior = componentInternalBehavior(
			internalAction(resourceDemand("Test"))
			);		
		/*
		 * @formatter:on
		 */
		ComponentInternalBehaviorMerger merger = new ComponentInternalBehaviorMerger(repository);
		merger.mergeTree(transaction, baseBehavior, otherBehavior, Collections.emptySet());
		assertThat(baseBehavior.getActions()).hasSize(1);
		assertThat(baseBehavior.getActions().get(0)).isInstanceOf(BranchAction.class);
	}

	@Test
	public void testMergeNoConflict() {
		/*
		 * @formatter:off
		 */
		ComponentInternalBehavior baseBehavior = componentInternalBehavior(
			internalAction(resourceDemand("Test"))
			);
		ComponentInternalBehavior otherBehavior = componentInternalBehavior(
			internalAction(resourceDemand("Test"))
			);		
		/*
		 * @formatter:on
		 */
		ComponentInternalBehaviorMerger merger = new ComponentInternalBehaviorMerger(repository);
		merger.mergeTree(transaction, baseBehavior, otherBehavior, Collections.emptySet());
		assertThat(baseBehavior.getActions()).hasSize(1);
		assertInternalAction(baseBehavior.getActions().get(0), "Test");
	}

	@Test
	public void testMergeBasicConflict() {
		/*
		 * @formatter:off
		 */
		ComponentInternalBehavior baseBehavior = componentInternalBehavior(
			internalAction(resourceDemand("Test1"))
			);
		ComponentInternalBehavior otherBehavior = componentInternalBehavior(
			internalAction(resourceDemand("Test2"))
			);		
		/*
		 * @formatter:on
		 */
		ComponentInternalBehaviorMerger merger = new ComponentInternalBehaviorMerger(repository);
		merger.mergeTree(transaction, baseBehavior, otherBehavior, Collections.emptySet());
		assertThat(baseBehavior.getActions()).hasSize(1);
		assertThat(baseBehavior.getActions().get(0)).isInstanceOf(BranchAction.class);
	}

	@Test
	public void testMergeBasicConflictWithCommonElement() {
		/*
		 * @formatter:off
		 */
		ComponentInternalBehavior baseBehavior = componentInternalBehavior(
			internalAction(resourceDemand("Common1")),
			internalAction(resourceDemand("Test1")),
			internalAction(resourceDemand("Common2"))
			);
		ComponentInternalBehavior otherBehavior = componentInternalBehavior(
			internalAction(resourceDemand("Common1")),
			internalAction(resourceDemand("Test2")),
			internalAction(resourceDemand("Common2"))
			);
		/*
		 * @formatter:on
		 */
		ComponentInternalBehaviorMerger merger = new ComponentInternalBehaviorMerger(repository);
		merger.mergeTree(transaction, baseBehavior, otherBehavior, Collections.emptySet());
		assertThat(baseBehavior.getActions()).hasSize(3);
		assertInternalAction(baseBehavior.getActions().get(0), "Common1");
		assertThat(baseBehavior.getActions().get(1)).isInstanceOf(BranchAction.class);
		assertInternalAction(baseBehavior.getActions().get(2), "Common2");
	}

	@Test
	public void testMergeBasicConflictWithTransitiveEdge1() {
		/*
		 * @formatter:off
		 */
		ComponentInternalBehavior baseBehavior = componentInternalBehavior(
			internalAction(resourceDemand("Common1")),
			internalAction(resourceDemand("Test1")),
			internalAction(resourceDemand("Common2"))
			);
		ComponentInternalBehavior otherBehavior = componentInternalBehavior(
			internalAction(resourceDemand("Common1")),
			internalAction(resourceDemand("Common2"))
			);
		/*
		 * @formatter:on
		 */
		ComponentInternalBehaviorMerger merger = new ComponentInternalBehaviorMerger(repository);
		merger.mergeTree(transaction, baseBehavior, otherBehavior, Collections.emptySet());
		assertThat(baseBehavior.getActions()).hasSize(3);
		assertInternalAction(baseBehavior.getActions().get(0), "Common1");
		assertThat(baseBehavior.getActions().get(1)).isInstanceOf(BranchAction.class);
		assertInternalAction(baseBehavior.getActions().get(2), "Common2");
	}

	@Test
	public void testMergeBasicConflictWithTransitiveEdge2() {
		/*
		 * @formatter:off
		 */
		ComponentInternalBehavior baseBehavior = componentInternalBehavior(
			internalAction(resourceDemand("Common1")),
			internalAction(resourceDemand("Common2"))
			);
		ComponentInternalBehavior otherBehavior = componentInternalBehavior(
			internalAction(resourceDemand("Common1")),
			internalAction(resourceDemand("Test1")),
			internalAction(resourceDemand("Common2"))
			);
		/*
		 * @formatter:on
		 */
		ComponentInternalBehaviorMerger merger = new ComponentInternalBehaviorMerger(repository);
		merger.mergeTree(transaction, baseBehavior, otherBehavior, Collections.emptySet());
		assertThat(baseBehavior.getActions()).hasSize(3);
		assertInternalAction(baseBehavior.getActions().get(0), "Common1");
		assertThat(baseBehavior.getActions().get(1)).isInstanceOf(BranchAction.class);
		assertInternalAction(baseBehavior.getActions().get(2), "Common2");
	}

	@Test
	public void testMergeComplexConflict() {
		/*
		 * @formatter:off
		 */
		ComponentInternalBehavior baseBehavior = componentInternalBehavior(
			internalAction(resourceDemand("A")),
			internalAction(resourceDemand("B")),
			internalAction(resourceDemand("C")),
			internalAction(resourceDemand("D"))
			);
		ComponentInternalBehavior otherBehavior1 = componentInternalBehavior(
			internalAction(resourceDemand("A")),
			internalAction(resourceDemand("C")),
			internalAction(resourceDemand("D"))
			);
		ComponentInternalBehavior otherBehavior2 = componentInternalBehavior(
			internalAction(resourceDemand("A")),
			internalAction(resourceDemand("B")),
			internalAction(resourceDemand("D"))
			);
		/*
		 * @formatter:on
		 */
		ComponentInternalBehaviorMerger merger = new ComponentInternalBehaviorMerger(repository);
		merger.mergeTree(transaction, baseBehavior, otherBehavior1, Collections.emptySet());
		merger.mergeTree(transaction, baseBehavior, otherBehavior2, Collections.emptySet());
		assertThat(baseBehavior.getActions()).hasSize(3);
		assertInternalAction(baseBehavior.getActions().get(0), "A");
		assertThat(baseBehavior.getActions().get(1)).isInstanceOf(BranchAction.class);
		assertInternalAction(baseBehavior.getActions().get(2), "D");
	}

	private static void assertInternalAction(AbstractAction action, String name) {
		assertThat(action).isInstanceOf(InternalAction.class);
		InternalAction casted = (InternalAction) action;
		assertThat(casted.getResourceDemand()).hasSize(1);
		assertThat(casted.getResourceDemand().get(0).getName()).isEqualTo(name);
	}

}
