package tools.descartes.prisma.core.scopes;

import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;

import edu.kit.ipd.descartes.mm.adaptationpoints.AdaptationPointDescriptions;
import edu.kit.ipd.descartes.mm.adaptationpoints.AdaptationpointsFactory;
import tools.descartes.prisma.core.ContainerPropertyNames;
import tools.descartes.prisma.core.ModelRepository;
import tools.descartes.prisma.core.Transaction;
import tools.descartes.prisma.core.adapter.ModelSkeletonAdapter;
import tools.descartes.prisma.core.exceptions.CommitException;
import tools.descartes.prisma.model.scope.Agent;
import tools.descartes.prisma.model.scope.DelegatedObject;
import tools.descartes.prisma.model.scope.ScopeConfiguration;
import tools.descartes.prisma.model.scope.ScopeFactory;
import tools.descartes.prisma.model.sensor.SensorFactory;
import tools.descartes.prisma.model.sensor.SensorRepository;
import tools.descartes.prisma.model.skeleton.ModelSkeleton;

public abstract class AgentScope extends Scope {

	private static final String SCOPE_CONFIGURATION_SUFFIX = ".scope";
	private static final String ADAPTATIONPOINTS_SUFFIX = ".adaptationpoints";
	private static final String SENSORS_SUFFIX = ".sensors";

	private final String type;
	private ScopeConfiguration configuration;
	private SensorRepository sensors;
	private AdaptationPointDescriptions adaptationPoints;
	private final SystemGlobalScope systemScope;
	
	public AgentScope(String name, String type, SystemGlobalScope systemScope, ModelRepository repository) {
		super(name, repository);
		this.type = type;
		this.systemScope = systemScope;
	}
	
	@Override
	protected String getBasePath() {
		return type;
	}

	public SystemGlobalScope getSystemScope() {
		return systemScope;
	}

	public SensorRepository getSensors() {
		if (sensors == null) {
			sensors = (SensorRepository) getModelRepository().getResource(getSensorsPath()).getContents().get(0);
		}
		return sensors;
	}

	private String getSensorsPath() {
		return type + "/" + getName() + "/" + getName() + SENSORS_SUFFIX;
	}
	
	public AdaptationPointDescriptions getAdaptationPoints() {
		if (adaptationPoints == null) {
			adaptationPoints = (AdaptationPointDescriptions) getModelRepository().getResource(getAdaptationPointsPath()).getContents().get(0);
		}
		return adaptationPoints;
	}
	
	private String getAdaptationPointsPath() {
		return type + "/" + getName() + "/" + getName() + ADAPTATIONPOINTS_SUFFIX;
	}
	
	public ScopeConfiguration getScopeConfiguration() {
		if (configuration == null) {
			configuration = (ScopeConfiguration) getModelRepository().getResource(getScopeConfigurationPath())
					.getContents().get(0);
		}
		return configuration;
	}

	private String getScopeConfigurationPath() {
		return type + "/" + getName() + "/" + getName() + SCOPE_CONFIGURATION_SUFFIX;
	}
	
	@Override
	public boolean create() {
		super.create();
		return super.create();
	}

	@Override
	protected void createResources(Transaction transaction) {
		Resource adaptationPointsResource = getModelRepository().getOrCreateResource(transaction,
				getAdaptationPointsPath());
		if (adaptationPointsResource.getContents().isEmpty()) {
			AdaptationPointDescriptions adaptationPoints = AdaptationpointsFactory.eINSTANCE
					.createAdaptationPointDescriptions();
			adaptationPoints.setName(getName());
			adaptationPointsResource.getContents().add(adaptationPoints);
		}
		
		Resource scopeResource = getModelRepository().getOrCreateResource(transaction, getScopeConfigurationPath());
		if (scopeResource.getContents().isEmpty()) {
			ScopeConfiguration configuration = ScopeFactory.eINSTANCE.createScopeConfiguration();
			scopeResource.getContents().add(configuration);
		}
		
		Resource sensorsResource = getModelRepository().getOrCreateResource(transaction, getSensorsPath());
		if (sensorsResource.getContents().isEmpty()) {
			SensorRepository sensors = SensorFactory.eINSTANCE.createSensorRepository();
			sensorsResource.getContents().add(sensors);
		}
	}

	public Agent getAgent(String id) {
		List<Agent> agents = getScopeConfiguration().getAgents();
		for (Agent curAgent : agents) {
			if (curAgent.getId().equals(id)) {
				return curAgent;
			}
		}
		return null;
	}

	public Agent addAgent(String id) {
		Agent agent = getAgent(id);
		if (agent == null) {
			try (Transaction transaction = getModelRepository().createTransaction()) {
				// Save this agent in the repository
				ScopeConfiguration scopeConfiguration = transaction.attach(getScopeConfiguration());
				transaction.lock(scopeConfiguration);
				agent = ScopeFactory.eINSTANCE.createAgent();
				agent.setId(id);
				scopeConfiguration.getAgents().add(agent);
				transaction.commit();
				
				return transaction.detach(agent);
			} catch (CommitException e) {
				e.printStackTrace();
			}
		}
		return agent;
	}

	public void removeAgent(Agent agent) {
		try (Transaction transaction = getModelRepository().createTransaction()) {
			ScopeConfiguration scopeConfiguration = transaction.attach(getScopeConfiguration());
			transaction.lock(scopeConfiguration);
			scopeConfiguration.getAgents().remove(agent);
			transaction.commit();
		} catch (CommitException e) {
			e.printStackTrace();
		}
	}

	public void registerElement(EObject element, Map<String, String> additionalProperties) {
		// Persist this delegated element in the delegations repository so
		// that agents that will arrive later can still access these elements
		try (Transaction curTransaction = getModelRepository().createTransaction()) {
			ScopeConfiguration scopeConfiguration = curTransaction.attach(getScopeConfiguration());
			curTransaction.lock(scopeConfiguration);

			// Search for existing delegation in the repository.
			for (DelegatedObject curElement : scopeConfiguration.getDelegatedObjects()) {
				if (curElement.getObject().equals(element)) {
					return;
				}
			}

			// Create a new delegation
			DelegatedObject newDelegation = ScopeFactory.eINSTANCE.createDelegatedObject();
			newDelegation.setObject(curTransaction.attach(element));
			newDelegation.getAdditionalProperties().putAll(additionalProperties);
			scopeConfiguration.getDelegatedObjects().add(newDelegation);
			curTransaction.commit();
		} catch (CommitException e) {
			e.printStackTrace();
		}
	}

	public void deregisterElement(EObject element) {
		// Remove the persisted element delegation from repository
		try (Transaction curTransaction = getModelRepository().createTransaction()) {
			ScopeConfiguration scopeConfiguration = curTransaction.attach(getScopeConfiguration());
			curTransaction.lock(scopeConfiguration);

			// Search for existing delegation in the repository and remove it.
			for (DelegatedObject curElement : scopeConfiguration.getDelegatedObjects()) {
				if (curElement.equals(element)) {
					scopeConfiguration.getDelegatedObjects().remove(curElement);
					break;
				}
			}
			curTransaction.commit();
		} catch (CommitException e) {
			e.printStackTrace();
		}
	}

	public void delegate(EObject element, Map<String, String> additionalProperties) {
		AgentScope scope = getDelegationScopes(element, additionalProperties);
		scope.registerElement(element, additionalProperties);
	}

	private AgentScope getDelegationScopes(EObject element, Map<String, String> additionalProperties) {
		String scopeName = additionalProperties.get(ContainerPropertyNames.CONTAINER_SCOPE);
		AgentScope scope = (AgentScope) getModelRepository().getScope(scopeName);
		if (scope != this) {
			scope.create();
		}
		return scope;
	}

	public void synchronizeModelSkeleton(Transaction transaction, ModelSkeleton skeleton) {
		List<ModelSkeletonAdapter> adapters = synchronizeModelSkeletonTree(transaction, skeleton);

		for (ModelSkeletonAdapter curAdapter : adapters) {
			curAdapter.synchronizeContents();
		}
	}

	protected List<ModelSkeletonAdapter> synchronizeModelSkeletonTree(Transaction transaction, ModelSkeleton skeleton) {
		List<ModelSkeletonAdapter> synchronizedAdapters = new LinkedList<>();
		if (skeleton.getAdaptationPoints() != null) {
			ModelSkeletonAdapter adapter = getModelRepository().adapt(skeleton.getAdaptationPoints(),
					ModelSkeletonAdapter.class);
			adapter.startTransaction(transaction);
			adapter.updateRepositoryObject(getAdaptationPoints());
			adapter.synchronizeTree();
			synchronizedAdapters.add(adapter);
		}
		if (skeleton.getSensors() != null) {
			ModelSkeletonAdapter adapter = getModelRepository().adapt(skeleton.getSensors(),
					ModelSkeletonAdapter.class);
			adapter.startTransaction(transaction);
			adapter.updateRepositoryObject(getSensors());
			adapter.synchronizeTree();
			synchronizedAdapters.add(adapter);
		}
		return synchronizedAdapters;
	}
}
